import os
from io import BytesIO
from typing import Annotated, Literal

from fastapi import APIRouter, Depends, File, Form, Header, Path, Response, UploadFile
from fastapi.params import Query
from fastapi.responses import StreamingResponse, UJSONResponse
from pydantic import EmailStr
from starlette import status

from fa_common import (
    BadRequestError,
    HTTPException,
    Message,
    get_settings,
    sizeof_fmt,
)
from fa_common import File as FileRef
from fa_common import logger as LOG
from fa_common.routes.user.models import UserDB
from fa_common.routes.user.service import get_current_app_user
from fa_common.storage import get_storage_client

from . import service
from .models import FileDB, UpdateFile

router = APIRouter()


async def valid_content_length(content_length: int = Header(...)):
    settings = get_settings()
    if content_length > settings.MAX_CONTENT_LENGTH:
        raise HTTPException(
            status_code=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE,
            error_name="File Upload Error",
            detail=f"This File was too large ({sizeof_fmt(content_length)}).\n"
            + f"The maximum file size is {sizeof_fmt(settings.MAX_CONTENT_LENGTH)}",
        )
    return content_length


@router.post(
    "",
    response_model=FileDB,
    responses={409: {"description": "File already exists"}, 413: {"description": "File too large"}},
)
async def upload_file(
    file: UploadFile = File(...),
    project_id: str | None = Form(None),
    sub_path: str = Form(None),
    tags: list[str] = Form([]),
    file_users: list[EmailStr] = Form([]),
    allow_duplicates: bool = Form(False),
    current_user: UserDB = Depends(get_current_app_user),
    file_size: int = Depends(valid_content_length),
):
    """Upload file to a project or user storage."""
    # permissions = lic_user[1]
    # if not permissions.allow_new_uploads:
    #     raise ForbiddenError(
    #         "Your current licence does not allow for uploading new datasets. "
    #         + "This is mostly likely caused by an expired Licence or Trial."
    #     )

    # if not allowed_file(file.filename):
    #     raise BadRequestError(
    #         detail="The file extension is not supported or there is something wrong with the filename.\n"
    #         + f"File Name: {file.filename} \n Supported extensions include: {ALLOWED_EXTENSIONS}"
    #     )
    settings = get_settings()
    file.file.seek(0, os.SEEK_END)
    file_size = file.file.tell()
    LOG.debug(f"File {file.filename} uploaded, size {file_size}")
    # max_file_size = permissions.upload_size_limit_bytes()
    max_file_size = settings.MAX_CONTENT_LENGTH
    if file_size > max_file_size:
        LOG.warning(f"File {file.filename} was too large at size {file_size} and was rejected")
        raise HTTPException(
            status_code=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE,
            detail=f"File {file.filename} was too large ({sizeof_fmt(file_size)}).\n"
            + f"The maximum file size is {sizeof_fmt(max_file_size)}",
        )
    file.file.seek(0)

    return await service.upload_file(
        file, current_user, project_id, sub_path, tags, file_users=file_users, allow_duplicates=allow_duplicates
    )


@router.get("", response_model=list[FileDB])
async def get_files(
    onlyMine=True,
    limit: int = 10,
    offset: int = 0,
    project: Annotated[list[str] | None, Query()] = None,
    path: str = "",
    startWith: str | None = None,
    exactName: str | None = None,
    mimeType: str | None = None,
    extension: str | None = None,
    current_user: UserDB = Depends(get_current_app_user),
):
    return await service.get_files_for_user(
        user=current_user,
        owner_only=onlyMine,
        offset=offset,
        limit=limit,
        project_ids=project,
        path=path,
        mime_type=mimeType,
        extension=extension,
        start_with=startWith,
        exact_name=exactName,
    )


@router.get(
    "/{file_id}",
    response_model=FileDB,
    responses={
        400: {"description": "Incorrectly formatted ID"},
        404: {"description": "File not found"},
        403: {"description": "Access Denied"},
    },
)
async def get_file_information(file_id: str, current_user: UserDB = Depends(get_current_app_user)):
    return await service.get_file(file_id, current_user)


async def file_streamer(file_bytes):
    for _i in range(10):
        yield b"some fake video bytes"


@router.get(
    "/{file_id}/download",
    responses={
        200: {
            "description": "Successful response, file is streamed to the client",
        },
        400: {"description": "Incorrectly formatted ID"},
        404: {"description": "File not found"},
        403: {"description": "Access Denied"},
    },
)
async def download_file(file_id: str, current_user: UserDB = Depends(get_current_app_user)):
    file = await service.get_file(file_id, current_user)
    file_bytes = await service.get_file_bytes(file)

    async def file_streamer(file_bytes: BytesIO):
        while True:
            chunk = file_bytes.read(4096)  # Read in chunks of 4KB
            if not chunk:
                break
            yield chunk

    return StreamingResponse(
        file_streamer(file_bytes),
        media_type=file.file_ref.content_type,
        headers={"Content-Disposition": f"attachment;filename={file.file_ref.name}"},
    )


@router.get(
    "/{file_id}/table",
    responses={
        400: {"description": "Incorrectly formatted ID or Invalid parameters and/or File type for table"},
        404: {"description": "File not found"},
        403: {"description": "Access Denied"},
        200: {
            "description": "Successful response",
            "content": {
                "application/json": {"schema": {"type": "object"}},
                "text/csv": {"schema": {"type": "string", "format": "binary"}},
            },
        },
    },
)
async def get_raw_table(
    file_id: str,
    offset: int = 0,
    limit: int = 100,
    seperator: str | None = None,
    sheet: str | None = None,
    accept: str = Header(None, alias="Accept"),
    current_user: UserDB = Depends(get_current_app_user),
):
    response_format: Literal["json", "csv"] = "json"
    if "application/json" in accept:
        # Handle JSON response
        response_format = "json"
    elif "text/csv" in accept:
        # Handle CSV response
        response_format = "csv"

    file = await service.get_file(file_id, current_user)

    response = await service.get_table_from_file(file, offset, limit, seperator, sheet, return_format=response_format)

    if response_format == "json":
        return UJSONResponse(response)
    else:
        return Response(response, media_type="text/csv")


@router.patch(
    "/{file_id}/file",
    responses={
        400: {"description": "Incorrectly formatted ID"},
        404: {"description": "File not found"},
        403: {"description": "Access Denied"},
        409: {"description": "File already exists"},
        413: {"description": "File too large"},
    },
    response_model=FileDB,
)
async def update_file(
    file_id: Annotated[str, Path(title="File Database ID")],
    file: UploadFile = File(...),
    allow_duplicates: bool = Form(False),
    current_user: UserDB = Depends(get_current_app_user),
    file_size: int = Depends(valid_content_length),
) -> FileDB:
    file_db = await service.get_file(file_id, current_user)

    settings = get_settings()
    file.file.seek(0, os.SEEK_END)
    file_size = file.file.tell()
    LOG.debug(f"File {file.filename} uploaded, size {file_size}")
    # max_file_size = permissions.upload_size_limit_bytes()
    max_file_size = settings.MAX_CONTENT_LENGTH
    if file_size > max_file_size:
        LOG.warning(f"File {file.filename} was too large at size {file_size} and was rejected")
        raise HTTPException(
            status_code=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE,
            detail=f"File {file.filename} was too large ({sizeof_fmt(file_size)}).\n"
            + f"The maximum file size is {sizeof_fmt(max_file_size)}",
        )
    file.file.seek(0)

    return await service.replace_file(file_db, file, allow_duplicates)


@router.get(
    "/{file_id}/presigned-url",
    response_model=FileDB,
    responses={
        400: {"description": "Incorrectly formatted ID"},
        404: {"description": "File not found"},
        403: {"description": "Access Denied"},
    },
)
async def get_file_link(file_id: str, expire_time_hours: int = 3, current_user=Depends(get_current_app_user)) -> FileRef:
    file = await service.get_file(file_id, current_user)

    client = get_storage_client()
    if file.file_ref.bucket is None or file.file_ref.id is None:
        raise BadRequestError("File reference is incomplete")
    return await client.create_temp_file_url(file.file_ref.bucket, file.file_ref.id, expire_time_hours)


# Delete file route
@router.delete(
    "/{file_id}",
    responses={
        400: {"description": "Incorrectly formatted ID"},
        404: {"description": "File not found"},
        403: {"description": "Access Denied"},
    },
    response_model=Message,
)
async def delete_file(file_id: str, current_user: UserDB = Depends(get_current_app_user)):
    file = await service.get_file(file_id, current_user)
    await service.delete_file(file)

    return Message(message=f"File {file.id} deleted successfully")


@router.patch(
    "/{file_id}",
    responses={
        400: {"description": "Incorrectly formatted ID"},
        404: {"description": "File not found"},
        403: {"description": "Access Denied"},
    },
    response_model=FileDB,
)
async def update_file_information(
    file_id: Annotated[str, Path(title="File Database ID")],
    update: UpdateFile,
    current_user: UserDB = Depends(get_current_app_user),
) -> FileDB:
    file = await service.get_file(file_id, current_user)
    # TODO: If changing project id the file should get moved to the new project
    return await service.update_file_metadata(file, update)
